# Copyright (c) 2024 Jordan Barrett & Aleksander Wojnarowicz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


__all__ = ["ABCDParams"]

from pydantic import (
    BaseModel,
    Field,
    field_validator,
)


class ABCDParams(BaseModel):
    gamma: float = Field(description="Power-law parameter for degrees, between 2 and 3", kw_only=True, default=2.5)
    delta: int = Field(description="Min degree", kw_only=True, default=5)
    zeta: float = Field(description="Parameter for max degree, between 0 and 1", kw_only=True, default=0.5)
    beta: float = Field(
        description="Power-law parameter for community sizes, between 1 and 2",
        kw_only=True,
        default=1.5,
    )
    s: int = Field(description="Min community size", kw_only=True, default=20)
    tau: float = Field(description="Parameter for max community size, between zeta and 1", kw_only=True, default=0.8)
    xi: float = Field(description="Noise parameter, between 0 and 1", kw_only=True, default=0.25)

    @field_validator("gamma")
    @classmethod
    def check_gamma(cls, v: float) -> float:
        if v < 2 or v > 3:
            raise ValueError("gamma must be between 2 and 3")
        return v

    @field_validator("zeta")
    @classmethod
    def check_zeta(cls, v: float) -> float:
        if v < 0 or v > 1:
            raise ValueError("zeta must be between 0 and 1")
        return v

    @field_validator("beta")
    @classmethod
    def check_beta(cls, v: float) -> float:
        if v < 1 or v > 2:
            raise ValueError("beta must be between 1 and 2")
        return v

    @field_validator("tau")
    @classmethod
    def check_tau(cls, v: float) -> float:
        if v < 0 or v > 1:
            raise ValueError("tau must be between 0 and 1")
        return v

    @field_validator("xi")
    @classmethod
    def check_xi(cls, v: float) -> float:
        if v < 0 or v > 1:
            raise ValueError("xi must be between 0 and 1")
        return v



# Copyright (c) 2024 Jordan Barrett & Aleksander Wojnarowicz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ["rand_round", "powerlaw_distribution", "get_community_color_map"]

import math
import random
from typing import TYPE_CHECKING

import numpy as np
from numpy.typing import NDArray

if TYPE_CHECKING:
    from abcd_graph.core.models import Community


def rand_round(x: float) -> int:
    p = x - math.floor(x)
    return int(math.floor(x) + 1) if random.uniform(0, 1) <= p else int(math.floor(x))


def powerlaw_distribution(choices: NDArray[np.int64], intensity: float) -> NDArray[np.float64]:
    dist: NDArray[np.float64] = (choices ** (-intensity)) / np.sum(choices ** (-intensity))
    return dist


def get_community_color_map(communities: list["Community"]) -> list[str]:
    import matplotlib.colors as colors  # type: ignore[import]

    colors_list = list(colors.BASE_COLORS.values())[: len(communities)]

    color_map = []

    for i, community in enumerate(communities):
        color = colors_list[i]
        color_map.extend([color] * len(community.vertices))

    return color_map



# Copyright (c) 2024 Jordan Barrett & Aleksander Wojnarowicz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ["build_communities", "build_degrees", "assign_degrees", "split_degrees", "build_community_sizes"]

from typing import Any

import numpy as np
from numpy.typing import NDArray

from abcd_graph.core.utils import (
    powerlaw_distribution,
    rand_round,
)
from abcd_graph.typing import (
    Communities,
    CommunitySizes,
    Degrees,
    DegreeSequence,
)


def build_degrees(n: int, gamma: float, delta: int, zeta: float) -> Degrees:
    max_degree = np.floor(n**zeta)
    avail = np.arange(delta, max_degree + 1)

    probabilities = powerlaw_distribution(avail, gamma)

    degrees = np.sort(np.random.choice(avail, size=n, p=probabilities))[::-1]

    if degrees.sum() % 2 == 1:
        degrees[0] += 1

    return degrees


def build_community_sizes(n: int, beta: float, s: int, tau: float) -> CommunitySizes:
    max_community_size = int(np.floor(n**tau))
    max_community_number = int(np.ceil(n / s))
    avail = np.arange(s, max_community_size + 1)

    probabilities = powerlaw_distribution(avail, beta)

    big_list: NDArray[np.int64] = np.random.choice(avail, size=max_community_number, p=probabilities)
    community_sizes: NDArray[np.int64] = np.zeros(max_community_number, dtype=np.int64)

    index = 0
    while community_sizes.sum() < n:
        community_sizes[index] = big_list[index]
        index += 1

    community_sizes = community_sizes[:index]
    excess = community_sizes.sum() - n
    if excess > 0:
        if (community_sizes[-1] - excess) >= s:
            community_sizes[-1] -= excess
        else:
            removed = community_sizes[-1]
            community_sizes = community_sizes[:-1]
            for i in range(removed - excess):
                community_sizes[i % len(community_sizes)] += 1
    return np.sort(community_sizes)[::-1]


def build_communities(community_sizes: CommunitySizes) -> Communities:
    communities = {}
    v_last = -1
    for i, c in enumerate(community_sizes):
        communities[i] = [v for v in range(v_last + 1, v_last + 1 + c)]
        v_last += c
    return communities


def assign_degrees(
    degrees: Degrees,
    communities: Communities,
    community_sizes: CommunitySizes,
    xi: float,
) -> dict[int, Any]:
    phi = 1 - np.sum(community_sizes**2) / (len(degrees) ** 2)
    deg = {}
    avail = 0
    already_chosen = set()

    lock = 0
    d_previous = degrees[0] + 1

    for i, d in enumerate(degrees):
        if (d < d_previous) and (lock < len(community_sizes)):
            threshold = d * (1 - xi * phi) + 1
            while community_sizes[lock] >= threshold:
                avail = communities[lock][-1]
                lock += 1
                if lock == len(community_sizes):
                    break

        v = np.random.choice(avail)
        while v in already_chosen:
            v = np.random.choice(avail)

        already_chosen.add(v)
        deg[v] = d

        if avail == len(degrees) - 1:
            still_not_chosen_set = set(range(len(degrees))) - already_chosen
            still_not_chosen: NDArray[np.int64] = np.array([v for v in still_not_chosen_set])
            degrees_remaining: NDArray[np.int64] = degrees[i + 1 :]  # noqa: E203

            np.random.shuffle(still_not_chosen)

            deg.update({label: degree for label, degree in zip(still_not_chosen, degrees_remaining)})
            return deg

        d_previous = d
    return deg


# TODO: naming degree list vs degree sequence (as dict)
def split_degrees(
    degrees: DegreeSequence,
    communities: Communities,
    xi: float,
) -> tuple[DegreeSequence, DegreeSequence]:
    deg_c = {v: rand_round((1 - xi) * degrees[v]) for v in degrees}
    for community in communities.values():
        if sum(deg_c[v] for v in community) % 2 == 0:
            continue

        v_max = _get_v_max(deg_c, community)
        deg_c[v_max] += 1
        if deg_c[v_max] > degrees[v_max]:
            deg_c[v_max] -= 2

    deg_b = {v: (degrees[v] - deg_c[v]) for v in degrees}
    return deg_c, deg_b


def _get_v_max(deg_c: dict[int, int], community: list[int]) -> int:
    deg_c_subset = {v: deg_c[v] for v in community}
    max_value = max(deg_c_subset.values())
    for elem, value in deg_c_subset.items():
        if value == max_value:
            return elem
    return community[0]



# Copyright (c) 2024 Jordan Barrett & Aleksander Wojnarowicz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = [
    "Graph",
]

from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
)

import numpy as np
from numpy.typing import NDArray

from abcd_graph.api.abcd_models import (
    Model,
    configuration_model,
)
from abcd_graph.core.build import (
    assign_degrees,
    build_communities,
    build_community_sizes,
    build_degrees,
    split_degrees,
)
from abcd_graph.core.exceptions import MalformedGraphException
from abcd_graph.core.models import ABCDGraph
from abcd_graph.core.utils import get_community_color_map
from abcd_graph.logger import construct_logger
from abcd_graph.utils import (
    require,
    require_graph_built,
)

if TYPE_CHECKING:
    from igraph import Graph as IGraph  # type: ignore[import]
    from networkx import Graph as NetworkXGraph  # type: ignore[import]

    from abcd_graph.api.abcd_params import ABCDParams


class Graph:
    def __init__(
        self,
        params: "ABCDParams",
        n: int = 1000,
        logger: bool = False,
    ) -> None:
        self.params = params
        self.n = n
        self.logger = construct_logger(logger)

        self._graph: Optional[ABCDGraph] = None

        self._model_used: Optional[Model] = None

    def reset(self) -> None:
        self._graph = None
        self._model_used = None

    @property
    @require_graph_built
    def summary(self) -> dict[str, Any]:
        assert self._graph is not None
        assert self._model_used is not None
        return {
            "number_of_nodes": self.n,
            "number_of_edges": self.num_edges,
            "number_of_communities": self.num_communities,
            "model": self._model_used.__name__,
            "is_proper_abcd": self.is_proper_abcd,
        }

    @property
    @require_graph_built
    def num_communities(self) -> int:
        assert self._graph is not None
        return self._graph.num_communities

    @property
    @require_graph_built
    def num_edges(self) -> int:
        assert self._graph is not None
        return len(self._graph.edges)

    @property
    def is_built(self) -> bool:
        return self._graph is not None

    @property
    @require_graph_built
    def is_proper_abcd(self) -> bool:
        assert self._graph is not None
        return self._graph.is_proper_abcd

    @property
    @require_graph_built
    def adj_matrix(self) -> NDArray[np.bool_]:
        if not self.is_proper_abcd:
            raise MalformedGraphException("Graph is not proper ABCD so the adjacency matrix is not accurate")

        assert self._graph is not None
        return self._graph.to_adj_matrix()

    @require("igraph")
    @require_graph_built
    def to_igraph(self) -> "IGraph":  # type: ignore[no-any-unimported]
        import igraph

        assert self._graph is not None
        return igraph.Graph(self._graph.edges)

    @require("networkx")
    @require_graph_built
    def to_networkx(self) -> "NetworkXGraph":  # type: ignore[no-any-unimported]
        import networkx as nx

        assert self._graph is not None

        graph = nx.Graph()

        graph.add_nodes_from(range(self.n))
        graph.add_edges_from(self._graph.edges)
        return graph

    @require("networkx")
    @require("matplotlib")
    @require_graph_built
    def draw_communities(self) -> None:
        if self._model_used is not None and self._model_used.__name__ != "configuration_model":
            raise NotImplementedError("Drawing communities is only supported for the configuration model")

        assert self._graph is not None

        if not self.is_proper_abcd:
            raise MalformedGraphException("Graph is not proper ABCD")

        import networkx as nx
        from matplotlib import pyplot as plt  # type: ignore[import]

        nx_g = self.to_networkx()

        color_map = get_community_color_map(communities=self._graph.communities)

        nx.draw(nx_g, node_color=color_map, with_labels=True, font_weight="bold")
        plt.show()

    def build(self, model: Optional[Model] = None) -> "Graph":
        model = model if model else configuration_model

        self._model_used = model

        degrees = build_degrees(
            self.n,
            self.params.gamma,
            self.params.delta,
            self.params.zeta,
        )

        self.logger.info("Building community sizes")

        community_sizes = build_community_sizes(
            self.n,
            self.params.beta,
            self.params.s,
            self.params.tau,
        )

        self.logger.info("Building communities")

        communities = build_communities(community_sizes)

        self.logger.info("Assigning degrees")

        deg = assign_degrees(degrees, communities, community_sizes, self.params.xi)

        self.logger.info("Splitting degrees")

        deg_c, deg_b = split_degrees(deg, communities, self.params.xi)

        self._graph = ABCDGraph(deg_b, deg_c)

        self.logger.info("Building community edges")
        self._graph.build_communities(communities, model)

        self.logger.info("Building background edges")
        self._graph.build_background_edges(model)

        self.logger.info("Resolving collisions")
        self._graph.combine_edges()

        self._graph.rewire_graph()

        return self



# Copyright (c) 2024 Jordan Barrett & Aleksander Wojnarowicz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from functools import wraps
from typing import (
    TYPE_CHECKING,
    Callable,
)

from typing_extensions import (
    ParamSpec,
    TypeVar,
)

if TYPE_CHECKING:
    from abcd_graph.abcd import Graph

P = ParamSpec("P")
R = TypeVar("R")


def require(package_name: str) -> Callable[[Callable[P, R]], Callable[P, R]]:
    def deco(func: Callable[P, R]) -> Callable[P, R]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            try:
                return func(*args, **kwargs)
            except ImportError as e:
                raise ImportError(
                    f"Package '{package_name}' is required to use '{func.__name__}'. Run "
                    f"`pip install abcd_graph[{package_name}]` to install it."
                ) from e

        return wrapper

    return deco


def require_graph_built(func: Callable[P, R]) -> Callable[P, R]:
    @wraps(func)
    def wrapper(self: "Graph", *args: P.args, **kwargs: P.kwargs) -> R:
        if not self.is_built:
            raise RuntimeError("Graph has not been built yet")

        return func(self, *args, **kwargs)  # type: ignore[arg-type]

    return wrapper  # type: ignore[return-value]



# Copyright (c) 2024 Jordan Barrett & Aleksander Wojnarowicz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = [
    "Model",
    "configuration_model",
    "chung_lu",
]

from typing import Protocol

import numpy as np
from numpy.typing import NDArray

from abcd_graph.typing import (
    Degrees,
    DegreeSequence,
)


class Model(Protocol):
    def __call__(self, degree_sequence: DegreeSequence) -> Degrees: ...

    @property
    def __name__(self) -> str: ...


def configuration_model(degree_sequence: DegreeSequence) -> Degrees:
    labels = list(degree_sequence.keys())
    counts = list(degree_sequence.values())

    vertices = np.repeat(labels, counts)

    np.random.shuffle(vertices)

    edges = np.array(vertices).reshape(-1, 2)

    return edges


def normalize(degrees: list[int]) -> NDArray[np.float64]:
    """Normalize the degree sequence."""
    degrees_array: NDArray[np.int64] = np.array(degrees)
    norm = degrees_array.sum()
    result: NDArray[np.float64] = np.divide(degrees_array, norm)
    return result


def chung_lu(degree_sequence: DegreeSequence) -> Degrees:
    """Generate a Chung-Lu random graph based on a given degree sequence."""
    nodes = list(degree_sequence.keys())
    degrees = list(degree_sequence.values())

    return np.random.choice(a=nodes, size=int(sum(degrees)), p=normalize(degrees)).reshape(-1, 2)



# Copyright (c) 2024 Jordan Barrett & Aleksander Wojnarowicz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = [
    "construct_logger",
]

import logging
import os
from abc import (
    ABC,
    abstractmethod,
)
from typing import Union

from typing_extensions import TypeAlias


class ABCDLogger(ABC):
    @abstractmethod
    def info(self, message: str) -> None:
        pass

    @abstractmethod
    def debug(self, message: str) -> None:
        pass

    @abstractmethod
    def warning(self, message: str) -> None:
        pass

    @abstractmethod
    def error(self, message: str) -> None:
        pass

    @abstractmethod
    def critical(self, message: str) -> None:
        pass


class NoOpLogger(ABCDLogger):
    def info(self, message: str) -> None:
        pass

    def debug(self, message: str) -> None:
        pass

    def warning(self, message: str) -> None:
        pass

    def error(self, message: str) -> None:
        pass

    def critical(self, message: str) -> None:
        pass


class StdOutLogger(ABCDLogger):
    NAME = "abcd-graph"

    def __init__(self) -> None:
        logging_level = int(os.getenv("ABCD_LOGGING_LEVEL", logging.INFO))
        logging.basicConfig(
            level=logging_level,
            format="[%(name)s] - %(asctime)s - %(levelname)s - %(message)s",
        )
        self.logging_level = logging_level
        self.logger = logging.getLogger(self.NAME)

    def info(self, message: str) -> None:
        self.logger.info(message)

    def debug(self, message: str) -> None:
        self.logger.debug(message)

    def warning(self, message: str) -> None:
        self.logger.warning(message)

    def error(self, message: str) -> None:
        self.logger.error(message)

    def critical(self, message: str) -> None:
        self.logger.critical(message)


LoggerType: TypeAlias = Union[ABCDLogger, bool]


def construct_logger(logger: bool) -> ABCDLogger:
    return StdOutLogger() if logger else NoOpLogger()



# Copyright (c) 2024 Jordan Barrett & Aleksander Wojnarowicz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__all__ = ["ABCDGraph", "Community"]

import abc
import random
from dataclasses import dataclass
from typing import (
    Any,
    Optional,
)

import numpy as np
from numpy.typing import NDArray

from abcd_graph.api.abcd_models import Model
from abcd_graph.typing import (
    Communities,
    DegreeSequence,
)


@dataclass
class Edge:
    __slots__ = ["v1", "v2"]

    v1: int
    v2: int

    def __post_init__(self) -> None:
        self.to_ordered()

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Edge):
            return NotImplemented
        return self.v1 == other.v1 and self.v2 == other.v2

    def __hash__(self) -> int:
        return hash((self.v1, self.v2))

    def to_ordered(self) -> None:
        if self.v1 < self.v2:
            self.v1, self.v2 = self.v2, self.v1

    @property
    def is_loop(self) -> bool:
        return self.v1 == self.v2


class AbstractGraph(abc.ABC):
    @property
    @abc.abstractmethod
    def adj_dict(self) -> dict[Edge, int]: ...


class AbstractCommunity(AbstractGraph):
    def __init__(self, edges: list[Edge]) -> None:
        self._adj_dict: dict[Edge, int] = {}
        self._bad_edges: list[Edge] = []

        for edge in edges:
            if edge.is_loop:
                self._bad_edges.append(edge)

            if edge in self.adj_dict:
                self._bad_edges.append(edge)
                self._adj_dict[edge] += 1
            else:
                self._adj_dict[edge] = 1

        self._edges = edges

    @property
    def vertices(self) -> list[int]:
        return list({edge.v1 for edge in self._edges} | {edge.v2 for edge in self._edges})

    @property
    def adj_dict(self) -> dict[Edge, int]:
        return self._adj_dict


class Community(AbstractCommunity):
    def push_to_background(self, edges: list[Edge], deg_b: DegreeSequence) -> None:
        for edge in edges:
            if edge.is_loop:
                for i in range(self.adj_dict[edge]):
                    self.adj_dict[edge] -= 1
                    if self.adj_dict[edge] == 0:
                        del self.adj_dict[edge]

                    deg_b[edge.v1] += 1
                    deg_b[edge.v2] += 1
            else:
                for i in range(self.adj_dict[edge] - 1):
                    self.adj_dict[edge] -= 1

                    deg_b[edge.v1] += 1
                    deg_b[edge.v2] += 1

    def rewire_community(self, deg_b: DegreeSequence) -> None:
        while len(self._bad_edges) > 0:
            for edge in self._bad_edges:
                other_edge = choose_other_edge(self.adj_dict, edge)
                rewire_edge(self.adj_dict, edge, other_edge)

            new_bad_edges = build_recycle_list(self.adj_dict)
            if len(new_bad_edges) >= len(self._bad_edges):
                self.push_to_background(new_bad_edges, deg_b)
                return
            else:
                self._bad_edges = new_bad_edges


class BackgroundGraph(AbstractCommunity):
    def __init__(self, edges: list[Edge]) -> None:
        super().__init__(edges)


class ABCDGraph(AbstractGraph):
    def __init__(self, deg_b: dict[int, int], deg_c: dict[int, int]) -> None:
        self.deg_b = deg_b
        self.deg_c = deg_c

        self.communities: list[Community] = []
        self.background_graph: Optional[BackgroundGraph] = None

        self._adj_dict: dict[Edge, int] = {}

    @property
    def adj_dict(self) -> dict[Edge, int]:
        return self._adj_dict

    def to_adj_matrix(self) -> NDArray[np.bool_]:
        adj_matrix = np.zeros((len(self.deg_b), len(self.deg_b)), dtype=bool)
        for edge in self._adj_dict:
            adj_matrix[edge.v1, edge.v2] = True
            adj_matrix[edge.v2, edge.v1] = True

        return adj_matrix

    @property
    def edges(self) -> list[tuple[int, int]]:
        return [(edge.v1, edge.v2) for edge in self._adj_dict]

    @property
    def is_proper_abcd(self) -> bool:
        return len(build_recycle_list(self._adj_dict)) == 0

    @property
    def num_communities(self) -> int:
        return len(self.communities)

    def build_communities(self, communities: Communities, model: Model) -> "ABCDGraph":
        for community in communities.values():
            community_edges = model({v: self.deg_c[v] for v in community})
            community_obj = Community([Edge(e[0], e[1]) for e in community_edges])
            community_obj.rewire_community(self.deg_b)

            assert len(build_recycle_list(community_obj.adj_dict)) == 0

            self.communities.append(community_obj)

        return self

    def build_background_edges(self, model: Model) -> "ABCDGraph":
        edges = [Edge(edge[0], edge[1]) for edge in model(self.deg_b)]
        self.background_graph = BackgroundGraph(edges)
        self._adj_dict = self.background_graph.adj_dict

        return self

    def combine_edges(self) -> "ABCDGraph":
        for community in self.communities:
            for edge, count in community.adj_dict.items():
                if edge in self._adj_dict:
                    self._adj_dict[edge] += count
                else:
                    self._adj_dict[edge] = count

        return self

    def rewire_graph(self) -> "ABCDGraph":
        bad_edges = build_recycle_list(self._adj_dict)

        while len(bad_edges) > 0:
            for edge in bad_edges:
                other_edge = choose_other_edge(self._adj_dict, edge)
                rewire_edge(self._adj_dict, edge, other_edge)

            bad_edges = build_recycle_list(self._adj_dict)

        return self


def build_recycle_list(adj_matrix: dict["Edge", int]) -> list["Edge"]:
    return [edge for edge in adj_matrix.keys() if adj_matrix[edge] > 1 or edge.is_loop]


def choose_other_edge(adj_matrix: dict["Edge", int], edge: "Edge") -> "Edge":
    edges = list(adj_matrix.keys())
    other_edge = random.choice(edges)
    while other_edge == edge:
        other_edge = random.choice(edges)

    return other_edge


def rewire_edge(adj_matrix: dict["Edge", int], edge: "Edge", other_edge: "Edge") -> None:
    if edge not in adj_matrix:
        return
    adj_matrix[edge] -= 1

    if adj_matrix[edge] == 0:
        del adj_matrix[edge]

    adj_matrix[other_edge] -= 1
    if adj_matrix[other_edge] == 0:
        del adj_matrix[other_edge]

    new_edge = Edge(edge.v1, other_edge.v1)

    if new_edge in adj_matrix:
        adj_matrix[new_edge] += 1
    else:
        adj_matrix[new_edge] = 1

    new_edge = Edge(edge.v2, other_edge.v2)

    if new_edge in adj_matrix:
        adj_matrix[new_edge] += 1
    else:
        adj_matrix[new_edge] = 1



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['abcd_graph', 'abcd_graph.api', 'abcd_graph.core']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.26.4,<2.0.0',
 'pydantic>=2.6.4,<3.0.0',
 'typing-extensions>=4.10.0,<5.0.0']

extras_require = \
{'all': ['pre-commit>=3.7.0,<4.0.0',
         'pytest>=8.1.1,<9.0.0',
         'pytest-cov>=5.0.0,<6.0.0',
         'networkx>=3.3,<4.0',
         'igraph',
         'matplotlib>=3.9.0,<4.0.0'],
 'dev': ['pre-commit>=3.7.0,<4.0.0',
         'pytest>=8.1.1,<9.0.0',
         'pytest-cov>=5.0.0,<6.0.0'],
 'igraph': ['igraph'],
 'matplotlib': ['matplotlib>=3.9.0,<4.0.0'],
 'networkx': ['networkx>=3.3,<4.0']}

setup_kwargs = {
    'name': 'abcd-graph',
    'version': '0.1.0',
    'description': '',
    'long_description': '# abcd-graph\nA python library for generating ABCD graphs.\n\n## Installation\n```bash\npip install abcd-graph\n```\n\n## Usage\n```python\nfrom abcd_graph import Graph, ABCDParams\n\nparams = ABCDParams()\ngraph = Graph(params).build()\n```\n\n### Parameters\n\n- `params`: An instance of `ABCDParams` class.\n- `n`: Number of nodes in the graph.\n- `logger` A boolean to enable or disable logging to the console. Default is `False` - no logs are shown.\n\n\n### Exporting\n\nThe graph object can be exported to `networkx.Graph` object using the `to_networkx` method.\n\n```python\nfrom abcd_graph import Graph, ABCDParams\n\nparams = ABCDParams()\ngraph = Graph(params).build().to_networkx()\n```\n\nThis requires the `networkx` library to be installed.\n```bash\npip install abcd-graph[networkx]\n```\n\nAnother option is an `igraph.Graph` object.\n\n```python\nfrom abcd_graph import Graph, ABCDParams\n\nparams = ABCDParams()\ngraph = Graph(params).build().to_igraph()\n```\n\nThis requires the `igraph` library to be installed.\n```bash\npip install abcd-graph[igraph]\n```\n\nFinally, the graph can be exported to a `numpy.ndarray` object that represents the `adjacency matrix`.\n\n```python\nfrom abcd_graph import Graph, ABCDParams\n\nparams = ABCDParams()\ngraph = Graph(params).build().adj_matrix\n```\n\n> [!IMPORTANT]\n> If the `build()` method is not run before calling any of the export methods, a `RuntimeError` will be raised.\n\n> [!NOTE]\n> The `numpy` array is of type `numpy.bool_`. If the graph was not properly generated (loops or multi-edges),\n> a `MalformedGraphError` will be raised.\n',
    'author': 'Your Name',
    'author_email': 'you@example.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
